"""Tests for :mod:`requests_gracedb.user_agent`."""
import pytest

from .. import __version__
from .. import Session

# FIXME: Python 2
pytest.importorskip('pytest_httpserver')


def test_user_agent(socket_enabled, httpserver):
    """Test that the User-Agent HTTP header is populated."""
    expected_user_agent = 'requests_gracedb/{}'.format(__version__)

    httpserver.expect_oneshot_request(
        '/', headers={'User-Agent': expected_user_agent}
    ).respond_with_data(
        'OK'
    )

    url = httpserver.url_for('/')
    client = Session(url)
    with httpserver.wait():
        client.get(url)



import os
from os import getuid
from six.moves.urllib.parse import urlparse

from safe_netrc import netrc

from .cert_reload import CertReloadingHTTPAdapter


def _find_x509_credentials():
    """Try to find a user's X509 certificate and key.

    Checks environment variables first, then expected location for default
    proxy.
    """
    proxy_file = os.environ.get('X509_USER_PROXY')
    cert_file = os.environ.get('X509_USER_CERT')
    key_file = os.environ.get('X509_USER_KEY')

    if cert_file and key_file:
        return cert_file, key_file

    if proxy_file:
        return proxy_file

    # Try default proxy
    proxy_file = os.path.join('/tmp', 'x509up_u{}'.format(getuid()))
    if os.path.exists(proxy_file):
        return proxy_file

    # Try default cert/key
    home_dir = os.environ.get('HOME')
    if home_dir:
        cert_file = os.path.join(home_dir, '.globus', 'usercert.pem')
        key_file = os.path.join(home_dir, '.globus', 'userkey.pem')

        if os.path.exists(cert_file) and os.path.exists(key_file):
            return cert_file, key_file

    return None


def _find_username_password(url):
    host = urlparse(url).hostname

    try:
        result = netrc().authenticators(host)
    except IOError:
        result = None

    if result is not None:
        username, _, password = result
        result = (username, password)

    return result


class SessionAuthMixin(object):
    """A mixin for :class:`requests.Session` to add support for all GraceDB
    authentication mechanisms.

    Parameters
    ----------
    url : str
        GraceDB Client URL.
    cert : str, tuple
        Client-side X.509 certificate. May be either a single filename
        if the certificate and private key are concatenated together, or
        a tuple of the filenames for the certificate and private key.
    username : str
        Username for basic auth.
    password : str
        Password for basic auth.
    force_noauth : bool, default=False
        If true, then do not use any authentication at all.
    fail_if_noauth : bool, default=False
        If true, then raise an exception if authentication credentials are
        not provided.
    cert_reload : bool, default=False
        If true, then automatically reload the client certificate before it
        expires.
    cert_reload_timeout : int, default=300
        Reload the certificate this many seconds before it expires.

    Notes
    -----
    When a new Session instance is created, the following sources of
    authentication are tried, in order:

    1.  If the :obj:`force_noauth` keyword argument is true, then perform no
        authentication at all.

    2.  If the :obj:`cert` keyword argument is provided, then use X.509 client
        certificate authentication.

    3.  If the :obj:`username` and :obj:`password` keyword arguments are
        provided, then use basic auth.

    4.  Look for a default X.509 client certificate in:

        a.  the environment variables :envvar:`X509_USER_CERT` and
            :envvar:`X509_USER_KEY`
        b.  the environment variable :envvar:`X509_USER_PROXY`
        c.  the file :file:`/tmp/x509up_u{UID}`, where :samp:`{UID}` is your
            numeric user ID
        d.  the files :file:`~/.globus/usercert.pem` and
            :file:`~/.globus/userkey.pem`

    5.  Read the netrc file [1]_ located at :file:`~/.netrc`, or at the path stored
        in the environment variable :envvar:`NETRC`, and look for a username
        and password matching the hostname in the URL.

    6.  If the :obj:`fail_if_noauth` keyword argument is true, and no
        authentication source was found, then raise a :class:`ValueError`.

    References
    ----------
    .. [1] The .netrc file.
           https://www.gnu.org/software/inetutils/manual/html_node/The-_002enetrc-file.html

    """  # noqa: E501

    def __init__(self, url=None, cert=None, username=None, password=None,
                 force_noauth=False, fail_if_noauth=False, cert_reload=False,
                 cert_reload_timeout=300, **kwargs):
        super(SessionAuthMixin, self).__init__(**kwargs)

        # Support for reloading client certificates
        if cert_reload:
            self.mount('https://', CertReloadingHTTPAdapter(
                cert_reload_timeout=cert_reload_timeout))

        # Argument validation
        if fail_if_noauth and force_noauth:
            raise ValueError(
                'Must not set both force_noauth and fail_if_noauth.')
        if (username is None) ^ (password is None):
            raise ValueError('Must provide username and password, or neither.')

        # FIXME: these should go into elif clauses below
        # (as in `elif default_cert := _find_x509_credentials():`)
        # in order to defer unnecessary I/O,  but this requires
        # the := operator, which requires Python 3.8.
        default_cert = _find_x509_credentials()
        default_basic_auth = _find_username_password(url)

        if force_noauth:
            pass
        elif cert is not None:
            self.cert = cert
        elif username is not None:
            self.auth = (username, password)
        elif default_cert is not None:
            self.cert = default_cert
        elif default_basic_auth is not None:
            self.auth = default_basic_auth
        elif fail_if_noauth:
            raise ValueError('No authentication credentials found.')



from __future__ import absolute_import
from os.path import basename
from mimetypes import guess_type
from sys import stdin

from requests.utils import guess_filename, to_key_val_list


def _guess_content_type(filename):
    return guess_type(filename)[0] or 'application/octet-stream'


def _guess_mime_type(key, val):
    if not isinstance(val, (tuple, list)):
        filename = guess_filename(val) or key
        filetype = _guess_content_type(filename)
        val = (filename, val, filetype)
    elif len(val) < 3 or val[2] is None:
        filename = val[0]
        filetype = _guess_content_type(filename)
        # val = (*val[:2], filetype, *val[3:])
        # FIXME: Python 2
        val = tuple(val[:2]) + (filetype,) + tuple(val[3:])
    return key, val


def _read_files(key, val):
    if isinstance(val, (tuple, list)) and (len(val) < 2 or val[1] is None):
        filename = val[0]
        with (stdin.buffer if filename == '-' else open(filename, 'rb')) as f:
            data = f.read()
        filename = basename(filename)
        # val = (filename, data, *val[2:])
        # FIXME: Python 2
        val = (filename, data) + tuple(val[2:])
    return key, val


def _prepare_file(key, val):
    key, val = _read_files(key, val)
    key, val = _guess_mime_type(key, val)
    return key, val


def _prepare_files(files):
    if files is not None:
        files = [_prepare_file(k, v) for k, v in to_key_val_list(files)]
    return files


class SessionFileMixin(object):
    """A mixin for :class:`requests.Session` to add features for file uploads.

    The :meth:`requests.Session.request` method takes a `files` argument which
    is a dictionary of `{fieldname: fileobject}`, where `fileobject` may be a
    file-like object or a tuple of 2-4 elements consisting of the filename,
    file content, MIME type, and any custom headers.

    This mixin adds the following features:

    * The MIME type is automatically guessed from the filename.
    * If the file content is None, then the filename is treated as a path, and
      the file is opened and read. If the filename is `-`, then the file
      content is read from stdin.
    """

    def request(
            self, method, url, params=None, data=None, headers=None,
            cookies=None, files=None, auth=None, timeout=None,
            allow_redirects=True, proxies=None, hooks=None, stream=None,
            verify=None, cert=None, json=None):
        return super(SessionFileMixin, self).request(
            method, url, params=params, data=data, headers=headers,
            cookies=cookies, files=_prepare_files(files), auth=auth,
            timeout=timeout, allow_redirects=True, proxies=proxies,
            hooks=hooks, stream=stream, verify=verify, cert=cert, json=json)



def _hook_raise_errors(response, *args, **kwargs):
    """Response hook to raise exception for any HTTP error (status >= 400)."""
    response.raise_for_status()


class SessionErrorMixin(object):
    """A mixin for :class:`requests.Session` to raise exceptions for HTTP
    errors.
    """

    def __init__(self, **kwargs):
        super(SessionErrorMixin, self).__init__(**kwargs)
        self.hooks['response'].append(_hook_raise_errors)



"""Tests for :mod:`requests_gracedb.file`."""
from __future__ import absolute_import
from mimetypes import guess_type
try:
    from unittest.mock import Mock
except ImportError:  # FIXME: Python 2
    from mock import Mock

import requests
import pytest

from .. import Session


@pytest.fixture
def mock_request(monkeypatch):
    """Mock up requests.Session base class methods."""
    mock = Mock()
    monkeypatch.setattr(requests.Session, 'request', mock)
    return mock


def test_filename_and_contents(mock_request, tmpdir):
    """Test handling of various styles of POSTed files."""
    # Different operating systems return different MIME types for *.xml files:
    # application/xml on macOS, text/xml on Linux.
    xml_mime_type, _ = guess_type('example.xml')

    client = Session('https://example.org/')
    filename = str(tmpdir / 'coinc.xml')
    filecontent = b'<!--example data-->'
    with open(filename, 'wb') as f:
        f.write(filecontent)
    file_expected = ('coinc.xml', filecontent, xml_mime_type)

    file_in = ('coinc.xml', filecontent)
    client.post('https://example.org/', files={'key': file_in})
    assert mock_request.call_args[1]['files'] == [('key', file_expected)]

    file_in = (filename, None)
    client.post('https://example.org/', files={'key': file_in})
    assert mock_request.call_args[1]['files'] == [('key', file_expected)]

    with open(filename, 'rb') as fileobj:
        file_in = fileobj
        file_expected = ('coinc.xml', fileobj, xml_mime_type)
        client.post('https://example.org/', files={'key': file_in})
        assert mock_request.call_args[1]['files'] == [('key', file_expected)]



"""Tests for :mod:`requests_gracedb.errors`."""
from __future__ import absolute_import
from requests.exceptions import HTTPError
import pytest

from .. import Session

# FIXME: Python 2
pytest.importorskip('pytest_httpserver')


def test_errors(socket_enabled, httpserver):
    """Test that HTTP 400 responses result in exceptions."""
    message = 'Tea time!'
    status = 418
    httpserver.expect_request('/').respond_with_data(message, status)

    url = httpserver.url_for('/')
    client = Session(url)
    with pytest.raises(HTTPError) as excinfo:
        client.get(url)
    exception = excinfo.value
    assert exception.response.status_code == status
    assert exception.response.reason == "I'M A TEAPOT"
    assert exception.response.text == message



"""Tests for :mod:`requests_gracedb.auth`."""
from __future__ import print_function
import os
import random
import stat

import pytest

from .. import Session


def set_rwx_user(fileobj):
    """Set correct permissions for X.509 or netrc files."""
    os.fchmod(fileobj.fileno(), stat.S_IRWXU)


@pytest.fixture
def x509_cert_and_key(tmpdir):
    """Generate (empty, dummy) X.509 public and private key files."""
    filenames = ('cert.pem', 'key.pem')
    filepaths = [str(tmpdir / filename) for filename in filenames]
    for filepath in filepaths:
        with open(filepath, 'wb') as f:
            set_rwx_user(f)
    return filepaths


@pytest.fixture
def x509_proxy(tmpdir):
    """Generate a (empty, dummy) X.509 proxy file."""
    filename = 'proxy.pem'
    filepath = str(tmpdir / filename)
    with open(filepath, 'wb') as f:
        set_rwx_user(f)
    return filepath


def test_noauth_invalid():
    """Test that setting force_noauth=fail_if_noauth=True is an error."""
    with pytest.raises(ValueError):
        Session('https://example.org/', force_noauth=True, fail_if_noauth=True)


def test_force_noauth():
    """Test force_noauth=True."""
    client = Session('https://example.org/', username='albert.einstein',
                     password='super-secret', force_noauth=True)
    assert client.auth is None
    assert client.cert is None


@pytest.mark.parametrize('username,password', [['albert.einstein', None],
                                               [None, 'super-secret']])
def test_basic_invalid(username, password):
    """Test that providing username or password, but not both, is an error."""
    with pytest.raises(ValueError):
        Session('https://example.org/', username=username, password=password)


def test_basic_explicit():
    """Test basic auth with explicitly provided username and password."""
    client = Session('https://example.org/', username='albert.einstein',
                     password='super-secret')
    assert client.auth == ('albert.einstein', 'super-secret')
    assert client.cert is None


def test_x509_explicit(x509_cert_and_key):
    """Test X.509 auth provided explicitly."""
    x509_cert, x509_key = x509_cert_and_key
    client = Session('https://example.org/', cert=(x509_cert, x509_key))
    assert client.auth is None
    assert client.cert == (x509_cert, x509_key)


def test_x509_default_cert_key(monkeypatch, x509_cert_and_key):
    """Test X.509 auth provided through X509_USER_CERT and X509_USER_KEY."""
    x509_cert, x509_key = x509_cert_and_key
    monkeypatch.setenv('X509_USER_CERT', x509_cert)
    monkeypatch.setenv('X509_USER_KEY', x509_key)
    client = Session('https://example.org/')
    assert client.auth is None
    assert client.cert == (x509_cert, x509_key)


def test_x509_default_proxy(monkeypatch, x509_proxy):
    """Test X.509 auth provided through X509_USER_CERT and X509_USER_KEY."""
    monkeypatch.delenv('X509_USER_CERT', raising=False)
    monkeypatch.delenv('X509_USER_KEY', raising=False)
    monkeypatch.setenv('X509_USER_PROXY', x509_proxy)
    client = Session('https://example.org/')
    assert client.auth is None
    assert client.cert == x509_proxy


@pytest.fixture
def x509up_exists(monkeypatch):
    """Make sure that the current UID has an X.509 proxy file.

    The default X.509 proxy file is at `/tmp/x509up_u{uid}`. Make sure
    that one exists. Create a temporary file with a name of that form
    and then monkeypatch the current uid.
    """
    while True:
        uid = random.randint(1000, 10000000)
        filename = '/tmp/x509up_u{}'.format(uid)
        # try:
        #     with open(filename, 'xb') as f:
        #         set_rwx_user(f)
        # except FileExistsError:
        #     continue
        # else:
        #     break
        # FIXME: racier Python 2 version
        if os.path.exists(filename):
            continue
        with open(filename, 'wb') as f:
            set_rwx_user(f)
            break
    monkeypatch.setattr('requests_gracedb.auth.getuid', lambda: uid)
    yield filename
    os.remove(filename)


@pytest.fixture
def x509up_does_not_exist(monkeypatch):
    """Make sure that the current UID does not have an X.509 proxy file.

    The default X.509 proxy file is at `/tmp/x509up_u{uid}`. Make sure
    that one does not exist. Find a random uid for which that file does
    not exist and then monkeypatch the current uid.
    """
    while True:
        uid = random.randint(1000, 10000000)
        filename = '/tmp/x509up_u{}'.format(uid)
        if not os.path.exists(filename):
            break
    monkeypatch.setattr('requests_gracedb.auth.getuid', lambda: uid)
    return filename


def test_x509_default_x509up(monkeypatch, tmpdir, x509up_exists):
    """Test X.509 auth provided through ~/.globus/user{cert,key}.pem."""
    monkeypatch.delenv('X509_USER_CERT', raising=False)
    monkeypatch.delenv('X509_USER_KEY', raising=False)
    monkeypatch.delenv('X509_USER_PROXY', raising=False)
    monkeypatch.setenv('HOME', str(tmpdir))
    client = Session('https://example.org/')
    assert client.auth is None
    assert client.cert == x509up_exists


def test_x509_default_globus(monkeypatch, tmpdir, x509up_does_not_exist):
    """Test X.509 auth provided through ~/.globus/user{cert,key}.pem."""
    monkeypatch.delenv('X509_USER_CERT', raising=False)
    monkeypatch.delenv('X509_USER_KEY', raising=False)
    monkeypatch.delenv('X509_USER_PROXY', raising=False)
    monkeypatch.setenv('HOME', str(tmpdir))
    os.mkdir(str(tmpdir / '.globus'))
    filenames = ['usercert.pem', 'userkey.pem']
    filepaths = [str(tmpdir / '.globus' / filename) for filename in filenames]
    for path in filepaths:
        with open(path, 'wb') as f:
            set_rwx_user(f)
    client = Session('https://example.org/')
    assert client.auth is None
    assert client.cert == tuple(filepaths)


def test_basic_default(monkeypatch, tmpdir, x509up_does_not_exist):
    """Test basic auth provided through a netrc file."""
    filename = str(tmpdir / 'netrc')
    with open(filename, 'w') as f:
        print('machine', 'example.org', 'login', 'albert.einstein',
              'password', 'super-secret', file=f)
        set_rwx_user(f)
    monkeypatch.setenv('NETRC', filename)
    monkeypatch.delenv('X509_USER_CERT', raising=False)
    monkeypatch.delenv('X509_USER_KEY', raising=False)
    monkeypatch.delenv('X509_USER_PROXY', raising=False)
    monkeypatch.setenv('HOME', str(tmpdir))
    client = Session('https://example.org/')
    assert client.auth == ('albert.einstein', 'super-secret')
    assert client.cert is None


def test_fail_if_noauth(monkeypatch, tmpdir, x509up_does_not_exist):
    """Test that an exception is raised if fail_if_noauth=True and no
    authentication source is available.
    """
    monkeypatch.setenv('NETRC', str(tmpdir / 'netrc'))
    monkeypatch.delenv('X509_USER_CERT', raising=False)
    monkeypatch.delenv('X509_USER_KEY', raising=False)
    monkeypatch.delenv('X509_USER_PROXY', raising=False)
    monkeypatch.setenv('HOME', str(tmpdir))
    client = Session('https://example.org/')
    assert client.auth is None
    assert client.cert is None
    with pytest.raises(ValueError):
        Session('https://example.org/', fail_if_noauth=True)



from __future__ import absolute_import

import requests.sessions

from .auth import SessionAuthMixin
from .errors import SessionErrorMixin
from .file import SessionFileMixin
from .user_agent import SessionUserAgentMixin

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

__all__ = ('Session',)


class Session(SessionAuthMixin,
              SessionErrorMixin,
              SessionFileMixin,
              SessionUserAgentMixin,
              requests.sessions.Session):
    """A :class:`requests.Session` subclass that adds behaviors that are common
    to ligo.org REST API services such as that of :doc:`GraceDB
    <gracedb:index>`.

    It adds the following behaviors to the session:

    * GraceDB-style authentication
      (see :class:`~requests_gracedb.auth.SessionAuthMixin`)

    * Raise exceptions based on HTTP status codes
      (see :class:`~requests_gracedb.error.SessionErrorMixin`)

    * Automatically load POSTed files from disk, automatically guess MIME types
      (see :class:`~requests_gracedb.file.SessionFileMixin`)

    * Add User-Agent string based on Python package name and version
      (see :class:`~requests_gracedb.user_agent.SessionUserAgentMixin`)
    """



"""Tests for :mod:`requests_gracedb.cert_reload`."""
from __future__ import absolute_import
from datetime import datetime
from ssl import SSLContext

from cryptography.x509 import (
    CertificateBuilder, DNSName, Name, NameAttribute, random_serial_number,
    SubjectAlternativeName)
from cryptography.x509.oid import NameOID
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric.rsa import generate_private_key
from cryptography.hazmat.primitives.serialization import (
    Encoding, NoEncryption, PrivateFormat)
from cryptography.hazmat.primitives.hashes import SHA256
import pytest
import six

from .. import Session

# FIXME: Python 2
pytest_httpserver = pytest.importorskip('pytest_httpserver')


@pytest.fixture
def backend():
    """Return an instance of the default cryptography backend."""
    return default_backend()


@pytest.fixture
def client_key(backend):
    """Generate client RSA key."""
    return generate_private_key(65537, 2048, backend)


@pytest.fixture
def server_key(backend):
    """Generate server RSA key."""
    return generate_private_key(65537, 2048, backend)


@pytest.fixture
def client_cert(client_key, backend):
    """Generate client certificate."""
    subject = issuer = Name([
        NameAttribute(NameOID.COMMON_NAME, six.u('example.org')),
        NameAttribute(NameOID.ORGANIZATION_NAME, six.u('Alice A. Client'))
    ])
    return CertificateBuilder().subject_name(
        subject
    ).issuer_name(
        issuer
    ).serial_number(
        random_serial_number()
    ).public_key(
        client_key.public_key()
    ).not_valid_before(
        datetime(3019, 1, 1)
    ).not_valid_after(
        datetime(3019, 1, 10)
    ).add_extension(
        SubjectAlternativeName([DNSName(six.u('localhost'))]),
        critical=False
    ).sign(
        client_key, SHA256(), backend
    )


@pytest.fixture
def server_cert(server_key, backend):
    """Generate server certificate."""
    subject = issuer = Name([
        NameAttribute(NameOID.COMMON_NAME, six.u('localhost')),
        NameAttribute(NameOID.ORGANIZATION_NAME, six.u('Bob B. Server'))
    ])
    return CertificateBuilder().subject_name(
        subject
    ).issuer_name(
        issuer
    ).serial_number(
        random_serial_number()
    ).public_key(
        server_key.public_key()
    ).not_valid_before(
        datetime(2008, 1, 1)
    ).not_valid_after(
        datetime(3020, 1, 1)
    ).add_extension(
        SubjectAlternativeName([DNSName(six.u('localhost'))]),
        critical=False
    ).sign(
        server_key, SHA256(), backend
    )


@pytest.fixture
def client_key_file(client_key, tmpdir):
    """Generate client key file."""
    filename = str(tmpdir / 'client_key.pem')
    with open(filename, 'wb') as f:
        f.write(client_key.private_bytes(
            Encoding.PEM, PrivateFormat.PKCS8, NoEncryption()))
    return filename


@pytest.fixture
def server_key_file(server_key, tmpdir):
    """Generate server key file."""
    filename = str(tmpdir / 'server_key.pem')
    with open(filename, 'wb') as f:
        f.write(server_key.private_bytes(
            Encoding.PEM, PrivateFormat.PKCS8, NoEncryption()))
    return filename


@pytest.fixture
def client_cert_file(client_cert, tmpdir):
    """Generate client certificate file."""
    filename = str(tmpdir / 'client_cert.pem')
    with open(filename, 'wb') as f:
        f.write(client_cert.public_bytes(Encoding.PEM))
    return filename


@pytest.fixture
def server_cert_file(server_cert, tmpdir):
    """Generate server certificate file."""
    filename = str(tmpdir / 'server_cert.pem')
    with open(filename, 'wb') as f:
        f.write(server_cert.public_bytes(Encoding.PEM))
    return filename


@pytest.fixture
def server(socket_enabled, server_cert_file, server_key_file):
    """Run test https server."""
    context = SSLContext()
    context.load_cert_chain(server_cert_file, server_key_file)
    with pytest_httpserver.HTTPServer(ssl_context=context) as server:
        server.expect_request('/').respond_with_json({'foo': 'bar'})
        yield server


@pytest.fixture
def client(server, client_cert_file, client_key_file, server_cert_file):
    """Create test client."""
    url = server.url_for('/')
    cert = (client_cert_file, client_key_file)
    with Session(url, cert=cert, cert_reload=True) as client:
        client.verify = server_cert_file
        yield client


def test_cert_reload(client, server, freezer):
    """Test reloading client X.509 certificates."""
    url = server.url_for('/')

    # Test 1: significantly before expiration time, still valid
    freezer.move_to('3019-01-02')
    assert client.get(url).json() == {'foo': 'bar'}
    pool1 = client.get_adapter(url=url).poolmanager.connection_from_url(url)
    conn1 = pool1.pool.queue[-1]

    # Test 2: > cert_reload_timeout seconds before expiration time, still valid
    freezer.move_to('3019-01-09T23:54:59')
    assert client.get(url).json() == {'foo': 'bar'}
    pool2 = client.get_adapter(url=url).poolmanager.connection_from_url(url)
    conn2 = pool2.pool.queue[-1]
    assert pool1 is pool2
    assert conn1 is conn2

    # Test 3: < cert_reload_timeout seconds before expiration time, invalid
    freezer.move_to('3019-01-10')
    assert client.get(url).json() == {'foo': 'bar'}
    pool3 = client.get_adapter(url=url).poolmanager.connection_from_url(url)
    conn3 = pool3.pool.queue[-1]
    assert pool1 is pool3
    assert conn1 is not conn3



from ._version import get_versions


class SessionUserAgentMixin(object):
    """A mixin for :class:`requests.Session` to add a User-Agent header."""

    def __init__(self, **kwargs):
        super(SessionUserAgentMixin, self).__init__(**kwargs)
        version = get_versions()['version']
        self.headers['User-Agent'] = '{}/{}'.format(__package__, version)



import sys

import versioneer
from setuptools import setup

setup_requires = ['setuptools >= 30.3.0']
if {'build_sphinx'}.intersection(sys.argv):
    setup_requires.append('sphinx')
if {'pytest', 'test', 'ptr'}.intersection(sys.argv):
    setup_requires.append('pytest-runner')

setup(setup_requires=setup_requires,
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass())



"""HTTPS adapter to close connections with expired client certificates."""
from __future__ import absolute_import
from datetime import datetime, timedelta
from functools import partial

from cryptography.hazmat.backends import default_backend
from cryptography.x509 import load_pem_x509_certificate
from urllib3.connection import HTTPSConnection
from urllib3.connectionpool import HTTPConnectionPool, HTTPSConnectionPool
from requests.adapters import HTTPAdapter

_backend = default_backend()


def load_x509_certificate(filename):
    """Load an X.509 certificate from a file.

    Parameters
    ----------
    filename : str
        The name of the certificate file.

    Returns
    -------
    cert : cryptography.x509.Certificate
        The parsed certificate.

    """
    with open(filename, 'rb') as f:
        data = f.read()
    return load_pem_x509_certificate(data, _backend)


class _CertReloadingHTTPSConnection(HTTPSConnection):

    def __init__(self, host, cert_reload_timeout=0, **kwargs):
        super(_CertReloadingHTTPSConnection, self).__init__(host, **kwargs)
        self._not_valid_after = datetime.max
        self._reload_timeout = timedelta(seconds=cert_reload_timeout)

    @property
    def cert_has_expired(self):
        expires = self._not_valid_after - datetime.utcnow()
        return expires <= self._reload_timeout

    def connect(self):
        if self.cert_file:
            cert = load_x509_certificate(self.cert_file)
            self._not_valid_after = cert.not_valid_after
        super(_CertReloadingHTTPSConnection, self).connect()


class _CertReloadingHTTPSConnectionPool(HTTPSConnectionPool):

    ConnectionCls = _CertReloadingHTTPSConnection

    def __init__(self, host, port=None, cert_reload_timeout=0, **kwargs):
        super(_CertReloadingHTTPSConnectionPool, self).__init__(
            host, port=port, **kwargs)
        self.conn_kw['cert_reload_timeout'] = cert_reload_timeout

    def _get_conn(self, timeout=None):
        while True:
            conn = super(_CertReloadingHTTPSConnectionPool, self)._get_conn(
                timeout)
            # Note: this loop is guaranteed to terminate because, even if the
            # pool is completely drained, when we create a new connection, its
            # `_not_valid_after` property is set to `datetime.max`, and the
            # condition below will evaulate to `True`.
            if not conn.cert_has_expired:
                return conn


class CertReloadingHTTPAdapter(HTTPAdapter):
    """A mixin for :class:`requests.Session` to automatically reload the client
    X.509 certificates if the version that is stored in the session is going to
    expire soon.

    Parameters
    ----------
    cert_reload_timeout : int
        Reload the certificate if it expires within this many seconds from now.

    """

    def __init__(self, cert_reload_timeout=0, **kwargs):
        super(CertReloadingHTTPAdapter, self).__init__(**kwargs)
        https_pool_cls = partial(
            _CertReloadingHTTPSConnectionPool,
            cert_reload_timeout=cert_reload_timeout)
        self.poolmanager.pool_classes_by_scheme = {
            'http': HTTPConnectionPool,
            'https': https_pool_cls}
